/*
 * Rubik's cube interface
 *
 * Copyright (C) 2007, 2008 Emmanuel Thome'
 * <Emmanuel dot Thome at normalesup dot org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef SIFT_HPP_
#define SIFT_HPP_

#include <istream>
#include <ostream>
#include <iostream>
#include <string>
#include <sstream>

#include <gmp.h>
#include <gmpxx.h>

#include "parser_base.hpp"

#define LEXGE2(X,Y,A,B) (X>A || (X == A && Y >= B))
#define LEXGE3(X,Y,Z,A,B,C) (X>A || (X == A && LEXGE2(Y,Z,B,C)))
#define GNUC_VERSION(X,Y,Z)     \
    (defined(__GNUC__) &&        \
(__GNUC__ == X && __GNUC_MINOR__ == Y && __GNUC_PATCHLEVEL__ == Z))
#define GNUC_VERSION_ATLEAST(X,Y,Z)     \
    (defined(__GNUC__) &&        \
LEXGE3(__GNUC__,__GNUC_MINOR__,__GNUC_PATCHLEVEL__,X,Y,Z))
#define GNUC_VERSION_ATMOST(X,Y,Z)     \
    (defined(__GNUC__) &&        \
LEXLE3(__GNUC__,__GNUC_MINOR__,__GNUC_PATCHLEVEL__,X,Y,Z))

#if  __INTEL_COMPILER || !GNUC_VERSION_ATLEAST(4,3,0)
#ifdef	__INTEL_COMPILER
#define STLEXT_NS       std
#else
#define	STLEXT_NS	__gnu_cxx
#endif
#include <ext/hash_set>
#define unordered_set hash_set
#else
#define	STLEXT_NS	std
#include <unordered_set>
#endif

template<int n>
struct hash_perm {
	size_t operator() (const perm<n>& __x) const {
                std::string s;
                std::ostringstream os;
                os << __x;
                using namespace STLEXT_NS;
                return hash<const char *>()(os.str().c_str());
	}
};


#if 0
/* This does not work. */
#ifdef	__INTEL_COMPILER
template<int n>
size_t hash_value(const perm<n> & p) {
	STLEXT_NS::hash<perm<n> > foo;
	return foo()(p);
}
#endif
#endif



template<int n>
class sift_table {
	typedef std::vector<int> iv_t;
	typedef std::pair<int,int> ipair_t;
	struct entry {
		perm<n> * s;
		iv_t w;
		entry() : s(NULL), w() {}
		entry(const entry& e) : s(new perm<n>(*(e.s))), w(e.w) {}
		~entry() { delete s; }
	};
	/****************************************************/
	entry tab[n][n];
	std::vector<ipair_t> todo_list;
	unsigned int done;
	STLEXT_NS::unordered_set<perm<n>,hash_perm<n> > considered;
	bool verb;
	std::vector<perm<n> > generators;
	iv_t generator_orders;
	std::vector<std::vector<bool> > commutation_table;
	public:
	sift_table() : done(0), verb(true), generators(0)
	{
		todo_list.reserve(n*n);
	}
	void quiet() { verb=false; }
	void loud() { verb=true; }
private:
	void back_end(perm<n>& s, ipair_t& p, iv_t& w, int i=0 /*,
			STLEXT_NS::hash_set<perm<n> >* pool=NULL */ ) const
	{
		// returns the place where the permutation would be
		// inserted (n,n if == id), together with its inverse
		// word.
		// Note that the input permutation is changed !
		// No insertion actually takes place.
		for(;i<n && s[i]==i;i++);
		if (i==n) { p.first=n; return; }

		/* if (pool) pool->insert(s); */

		int j=s[i];
		p=std::make_pair(i,j);
		if (tab[i][j].s==NULL) {
			if (!commutation_table.empty()) {
				compress_word(w);
			}
			return;
		}
		if (verb) std::cout << "Cell " << p << " busy, descending ; ";
		perm<n> t = tab[i][j].s->inverse();
		s=t*s;
		const iv_t& tijw (tab[i][j].w);
		typedef iv_t::const_reverse_iterator vicri_t;
		for(vicri_t ptr=tijw.rbegin();ptr!=tijw.rend();ptr++) {
			w.push_back(-*ptr);
		}
		if (verb) std::cout << "perm: " << s << " ; ";
		if (verb) std::cout << "word: " << w << "\n";
		return back_end(s,p,w,i+1 /*,pool */);
	}
	/* OK, this uglifies the code. But here we introduce an
	 * optimization which _rocks_ ! Gains a 10x factor on the final
	 * word size !
	 */
	void back_end_push(perm<n>& s, ipair_t& p, iv_t& w, int i=0)
	{
		for(;i<n && s[i]==i;i++);
		if (i==n) { p.first=n; return; }

		considered.insert(s);

		int j=s[i];
		p=std::make_pair(i,j);
		if (tab[i][j].s==NULL) {
			if (!commutation_table.empty()) {
				compress_word(w);
			}
			return;
		}
		if (verb) std::cout << "Cell " << p << " busy, descending ; ";
		perm<n> t = tab[i][j].s->inverse();
		iv_t tijw = tab[i][j].w;
		{
			/* XXX ; here is the trick ; pretty obvious, in
			 * fact.
			 */
			if (tijw.size() > w.size()) {
				if (verb) std::cout << "YAHOO " << w.size()
					<< "<" << tijw.size() << "\n";
				// replace b by a ; proceed with b^-1a
				delete tab[i][j].s;
				tab[i][j].s = new perm<n>(s);
				tab[i][j].w = w;
			}
		}
		s=t*s;
		typedef iv_t::const_reverse_iterator vicri_t;
		for(vicri_t ptr=vicri_t(tijw.rbegin());ptr!=vicri_t(tijw.rend());ptr++) {
			w.push_back(-*ptr);
		}
		if (verb) std::cout << "perm: " << s << " ; ";
		if (verb) std::cout << "word: " << w << "\n";
		return back_end_push(s,p,w,i+1);
	}
public:
	bool push(const perm<n>& os, const iv_t& ow, int i=0) {
		perm<n> s=os;
		iv_t w=ow;

		ipair_t p;
		//back_end(s,p,w,i,&considered);
		back_end_push(s,p,w,i);
		if (p.first==n) return false;

		if (verb) std::cout << "Filling cell " << p << "\n";
		tab[p.first][p.second].s=new perm<n> (s);
		tab[p.first][p.second].w=w;
		todo_list.push_back(p);

		return true;
	}
	bool inverse(const perm<n>& os, iv_t& w) {
		perm<n> s=os;
		w.clear();
		ipair_t p;
		back_end(s,p,w);
		ASSERT(!commutation_table.empty());
		compress_word(w);
		return (p.first==n);
	}
	void push(const perm<n>& s, int k) { push(s,iv_t(1,k),0); }
	void generator(const perm<n>& s) {
		generators.push_back(s);
		generator_orders.push_back(s.order());
		int k=generators.size();

		/* Should we take the powers of the generators as well ?
		 * Experience shows that this improves only very
		 * marginally the results.
		 */
#ifdef	WITH_POWERS
		int o=s.order();
#endif
		{
#ifdef	WITH_POWERS
			int l=o/2;
#else
			int l=1;
#endif
			perm<n> t;// = id by default
			iv_t w;
			for(int i=0;i<l;i++) {
				t=t*s;
				w.push_back(k);
				push(t,w);
			}
		}
		{
#ifdef	WITH_POWERS
			int l=o/2;
#else
			int l=1;
#endif
			perm<n> t;// = id by default
			iv_t w;
			perm<n> si=s.inverse();
			for(int i=0;i<l;i++){
				t=t*si;
				w.push_back(-k);
				push(t,w);
			}
		}
	}
	void fix_generators() {
		const std::vector<perm<n> >& g(generators);
		std::vector<std::vector<bool> >& c(commutation_table);
		c.assign(g.size(),std::vector<bool>(g.size(),false));
		for(unsigned int i=0;i<g.size();i++) {
			c[i][i]=true;
			for(unsigned int j=i+1;j<g.size();j++) {
				if (g[i]*g[j] == g[j]*g[i]) {
					if (verb)
						std::cout << "Gens " << i << " and "
						<< j << " commute\n";
					c[i][j]=c[j][i]=true;
				}
			}
		}
	}
	/* compresses in place */
	void compress_word(iv_t& w) const {
		unsigned int os=w.size();
		typedef iv_t::iterator vii_t;
		vii_t first, last;
		first=last=w.begin();
		const unsigned int& ng = generators.size();
		const std::vector<std::vector<bool> >& c(commutation_table);
		iv_t val;
		unsigned int e;
		for(;first!=w.end();) {
			ASSERT(first==last);
			val.assign(ng,0);
			e=ABS(*first)-1;
			val[e]=(*first>0)?1:-1;
			for(;first!=w.begin();first--) {
				int f = ABS(*(first-1))-1;
				if (!c[e][f]) break;
				val[f]+=(*(first-1)>0)?1:-1;
			}
			for(last++;last!=w.end();last++) {
				int f = ABS(*last)-1;
				if (!c[e][f]) break;
				val[f]+=(*last>0)?1:-1;
			}
			int p=0;
			for(e=0;e<ng;e++) {
				const int& k=generator_orders[e];

				val[e] = val[e] % k;
				if (val[e]==-k/2) val[e]+=k;
				if (ABS(val[e]+k) < ABS(val[e])) val[e]+=k;
				if (ABS(val[e]-k) < ABS(val[e])) val[e]-=k;
				p+=ABS(val[e]);
			}
			ASSERT(p<=(last-first));
			/*
			if (p<(last-first))
			std::cout << "Replacing sequence of length "
				<< (last-first) << " at position "
				<< (first-w.begin()) << " by sequence"
				<< " of length " << p << "\n";
				*/
			for(e=0;e<ng;e++) {
				if (val[e]==0) continue;
				int f=(val[e]>0)?1:-1;
				f*=(1+e);
				int k=ABS(val[e]);
				for(;k--;*first++=f);
			}
			w.erase(first,last);
			last=first;
		}
		if (w.size() < os) {
			if (verb) std::cout << "Compress down from "
				<< os << " to " << w.size() << "\n";
		}
	}
	bool finished() const { return done==todo_list.size(); }
	void apply() {
		if (verb) std::cout << "BEGIN NEW PASS\n";
		unsigned int fence=todo_list.size();
		typedef std::vector<ipair_t>::const_iterator vpici_t;
		vpici_t ws=todo_list.begin()+done;
		vpici_t we=todo_list.end();
		for(vpici_t doing=ws;doing!=we;doing++) {
			for(vpici_t other=todo_list.begin();other!=we;other++) {
				// consider the product (*doing) * (*other)
				// and (*other) * (*doing) as well.
				iv_t w;
				std::back_insert_iterator<iv_t > wb(w);

				int i1,j1,i2,j2;
				perm<n> s;

				i1=doing->first;j1=doing->second;
				i2=other->first;j2=other->second;

				const entry& ea(tab[i1][j1]);
				const entry& eb(tab[i2][j2]);
				
				if (verb) std::cout << "Prod: "
					<< *other << "*" << *doing << " ; ";

				s=(*(eb.s)) * (*(ea.s));
				if (considered.find(s)==considered.end()) {
					copy(ea.w.begin(),ea.w.end(),wb);
					copy(eb.w.begin(),eb.w.end(),wb);
					if (verb) std::cout << "perm: " << s << " ; ";
					if (verb) std::cout << "word: " << w << "\n";
					push(s,w);
				} else {
					if (verb) std::cout << "done already\n";
				}

				if (*doing == *other) continue;

				w.clear();

				if (verb) std::cout << "Prod: "
					<< *doing << "*" << *other << " ; ";

				s=(*(ea.s)) * (*(eb.s));
				if (considered.find(s)==considered.end()) {
					copy(eb.w.begin(),eb.w.end(),wb);
					copy(ea.w.begin(),ea.w.end(),wb);
					if (verb) std::cout << "perm: " << s << " ; ";
					if (verb) std::cout << "word: " << w << "\n";
					push(s,w);
				} else {
					if (verb) std::cout << "done already\n";
				}
				w.clear();
			}
		}
		done=fence;
	}
public:
	class sift_parser : public perm<n>::perm_parser {
		typedef typename perm<n>::perm_parser super_t;
		void want_perm(perm<n>& g) {
			/*
			std::istringstream is(string_version());
			std::cout << "Before want_perm, s=" << is.str() << "\n";
			std::cout << "Read " << is.tellg() << " chars\n";
			if (!(is >> g)) {
				error("want a permutation");
			}
			std::cout << "After want_perm, s=" << is.str() << "\n";
			std::cout << "Read " << is.tellg() << " chars\n";
			discard(is.tellg());
			*/
			require(super_t::do_parse(g));
		}
		void want_int_word(iv_t& w) {
			w.clear();
			parser_base::want('[');
			for(;parser_base::numlen_sign();) {
				w.push_back(parser_base::want_si());
			}
			parser_base::want(']');
		}
		public:
		bool is_comment() { return parser_base::ptry("// "); }
		sift_parser(const char * obuf, unsigned int osize) :
				super_t(obuf,osize) {}
		bool parse_gen(perm<n>& res) {
			if (!parser_base::ptry("GEN"))
				return false;
			parser_base::want_ui();
			parser_base::want(':');
			want_perm(res);
			return true;
		}
		bool parse_entry(sift_table<n>& res) {
			if (!parser_base::ptry("AT"))
				return false; 
			int i,j;
			parser_base::want('(');
			i=parser_base::want_ui();
			parser_base::want(',');
			j=parser_base::want_ui();
			parser_base::want(')');
			parser_base::want(':');
			perm<n> g;
			want_perm(g);
			parser_base::want(';');
			iv_t w;
			want_int_word(w);
			res.push(g,w,i);
			return true;
		}
		bool do_parse(sift_table<n>& res) {
			perm<n> g;
			if (is_comment())
				return true;
			if (res.commutation_table.empty()) {
				if (parse_gen(g)) {
					res.generator(g);
					return true;
				}
				res.fix_generators();
			}
			return parse_entry(res);
		}
	};
public:
	std::istream& operator>>(std::istream& is) {
		const unsigned int buffer_size=parser_base::buffer_size;
		char    buffer[buffer_size];
		for(;;) {
			is.getline(buffer,buffer_size);
			if (is.gcount()==0) break;
			sift_parser j(buffer,is.gcount());
			if (!j.do_parse(*this)) {
				is.setstate(std::ios::failbit);
			}
		}
		return is;
	}
	void profile(std::ostream& os) const {
		for(int i=0;i<n;i++) {
			os << "// 0";
			for(int j=i+1;j<n;j++) {
				if (tab[i][j].s==NULL) os << " -1";
				else os << " " << tab[i][j].w.size();
			}
			os << "\n";
		}
	}
	std::ostream& operator<<(std::ostream& os) const {
		for(unsigned int i=0;i<generators.size();i++) {
			os << "GEN " << (1+i) << ": " << generators[i] << "\n";
		}
		for(int i=0;i<n;i++) {
			os << "// LINE " << i << "\n";
			for(int j=i+1;j<n;j++) {
				if (tab[i][j].s==NULL) continue;
				os << "AT (" << i
					<< "," << j << ") : "
					<< *(tab[i][j].s) << " ; ";
				os << tab[i][j].w << "\n";
			}
		}
		profile(os);
		return os;
	}
	mpz_class group_order() const {
		mpz_class o=1;
		for(int i=0;i<n;i++) {
			int k=1;
			for(int j=i+1;j<n;j++) {
				if (tab[i][j].s==NULL) continue;
				k++;
			}
			o*=k;
		}
		return o;
	}
	perm<n> random(gmp_randstate_t toto) const {
		perm<n> r;
		mpz_class bob;
		mpz_class fred;
		for(int i=0;i<n;i++) {
			std::vector<perm<n> > tmp;
			tmp.push_back(perm<n>());
			for(int j=i+1;j<n;j++) {
				if (tab[i][j].s==NULL) continue;
				tmp.push_back(*(tab[i][j].s));
			}
			fred=tmp.size();
			mpz_urandomm(bob.get_mpz_t(),toto,fred.get_mpz_t());
			r*=tmp[bob.get_ui()];
		}
		return r;
	}
};

template<int n>
std::ostream& operator<<(std::ostream& os, const sift_table<n>& a)
{
	return a.operator<<(os);
}

template<int n>
std::istream& operator>>(std::istream& is, sift_table<n> & a)
{
        return a.operator>>(is);
}

#endif	/* SIFT_HPP_ */
