/*
 * Rubik's cube interface
 *
 * Copyright (C) 2007, 2008 Emmanuel Thome'
 * <Emmanuel dot Thome at normalesup dot org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <fstream>
#include <iostream>
#include <vector>
#include <list>
#include <queue>
#include <iterator>
#include <cstring>

#include "manu.h"
#include "parser_base.hpp"
#include "util.hpp"
#include "perm.hpp"
#include "sift.hpp"
#include "input_sift_simple.hpp"

#define	NGENS	(sizeof(SOURCE)/sizeof(SOURCE[0]))

int main(int argc, char **argv)
{
	sift_table<DIM> t;

	if (argc>1 && strcmp(argv[1],"--do-sift")==0) {
		perm<DIM> faces[NGENS];


		for(unsigned int i=0;i<NGENS;i++) {
			std::istringstream is(SOURCE[i]);
			is >> faces[i];
			std::cout << faces[i] << "\n";
			t.generator(faces[i]);
		}
		t.fix_generators();

		int pass;
		for(pass=0;!t.finished();pass++) {
			std::cout << "Sifting, pass " << pass << "\n";
			t.apply();
		}

		std::cout << "Finished sifting in "<< pass <<" passes\n";

		std::ofstream data_file;
		data_file.open("simple.sft", std::ios_base::out);
		data_file << t ;
		data_file.close();
	} else {
		std::ifstream data_file;
		data_file.open("simple.sft", std::ios_base::in);
		try {
			data_file >> t;
		} catch (parse_error x) {
			std::cerr << "Parse error reading file : "
				<< x.msg << "\n";
		}
	}

	gmp_randstate_t toto;
	gmp_randinit_default(toto);

	std::cout << "Group order: " << t.group_order() << "\n";

	for(int k=0;k<100;k++)
		std::cout << "Random permutation: " << t.random(toto) << "\n";

	gmp_randclear(toto);
	t.profile(std::cout);
}
